;;; stout.el
;;; * Outline mode starts here.

;; Copyright (C) 1998 Tom Breton

;; Time-stamp: <1998-12-05 19:16:39 Tehom>
;; Author: Tom Breton <tob@world.std.com>
;; Maintainer: Tom Breton until he gets overloaded <tob@world.std.com>
;; Created:  3 Oct 1998
;; Version: 0.2
;; Keywords: outlines

;;
;; Purpose of this package:
;;
;; To allow outlines to be more powerfully structured.
;;
;; Installation instructions:
;;
;; Put stout.el and local-vars.el somewhere in your load path.
;;
;; Add to your .emacs:
;;  (autoload 'stout-mode "stout" 
;;    "Major/Minor mode for managing structured outlines" t)
;;
;; Read stout.doc.  stout.doc should have been included with the stout
;; distribution.
;;
;; 
;;
;; Known bugs:
;;
;; LCD Archive Entry:
;; 
;; stout.el||Tehom@localhost
;; |Structured outline mode
;; Tue Oct 20 18:45:43 1998|0.1||
;;
;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2 of
;; the License, or (at your option) any later version.

;; This program is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY; without even the implied
;; warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
;; PURPOSE.  See the GNU General Public License for more details.

;; You should have received a copy of the GNU General Public
;; License along with this program; if not, write to the Free
;; Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
;; MA 02111-1307 USA

;;; * Commentary:

;; "stout" stands for "ST_ructured OUT_lining"
;; The change log for this file is in stout.changes.
;; The documentation for this file is in stout.doc.

;;; * Code:

;;; ** Required packages

( require 'outline )
( require 'cl)
( autoload 'using-local-vars "local-vars" "" nil 'macro)

;;; ** Variables

(defgroup stout nil "Stout structured outline mode"
  :group 'outlines)

(defcustom stout-propertize-text t
  "*Should inserted structured text be fontified?"
  :type 'boolean
  :group 'stout)


;;; *** Structuring objects

(defvar stout-structuring-objects nil
  "List of available structuring objects.  

This variable is made buffer-local by stout-mode."
  )

(defvar stout-views nil
  "List of available views.  

This variable is made buffer-local by stout-mode."
  )


;;; *** Regular expressions to control searching.


;;Important: outline-regexp must begin something that distinguishes
;;the start of an expression, such as "^".
;;
;;outline.el's default value has no way of indicating the start of a
;;pattern, so we set a better default.
;;
;;We want to change the default outline.el gives, because it has no
;;way of recognizing the beginning of a string (outline handles it
;;with a hack).  We don't want to change outline-regexp if something
;;else has set it -- say, emacs-lisp-mode.  There is no good way of
;;distinguishing the two.  However, since we're talking about the
;;global default, it should not be as crucial as outline.el thinks.

;;^L-header nodes should be stout nodes as well.
(setq-default outline-regexp "^[*\^L]+")

(defvar stout-value-end-regxep "$" )

;;; *** Insertion format

(defvar stout-insert-format
  '(
     header-string
     " "
     handle-type-string
     " "
     handle-name-string
     "\n"
     body-string
     "\n"
     )
  "The format list for insertion.  

This variable shouldn't be changed unless you are deliberately
changing the outline format, in which case it's up to you to make all
the formatting match, which would require serious rewriting."
  )

;;; *** Variables indicating which functions to call.

(defvar stout-make-header-string 'stout-default-make-header-string 
  "The function that makes a header string suitable for a given depth.

This variable shouldn't be changed unless you are deliberately
changing the outline format, in which case it's up to you to make all
the formatting match."
  )


;;; *** Property lists.

;;Properties for various regions of a node.  
;;
;;To make any of these region invisible, set the respective list to be
;;or include '(invisible outline).  Header shouldn't be made invisible
;;or it will confuse outline.el, because outline will think that means
;;the whole node is invisible.
;;

(defvar stout-header-text-props nil 
  "Properties a node's header text should have by default.
This should never include the property 'invisible.")

(defvar stout-type-text-props '(invisible outline) 
  "Properties a node's type text should have by default.")

(defvar stout-name-text-props nil 
  "Properties a node's name text should have by default.")

(defvar stout-body-text-props nil 
  "Properties a node's body text should have by default.")


;;; ** Structures

;;; *** Notes

;;Note: :short-doc and :documentation aren't real slot options, but I hope
;;one day they will be.  Slots need docs too.

;;:short-doc should be a single line, suitable for use in 
;;  (format "The slot \"%s\" contains %s."  slot-name short-doc).

;;There is no length limit on documentation.


;;; *** Handle structures.

;;This is separate because we needed handle structures to know the
;;name of the type we're writing.

(defstruct stout-handle-type 
  "The structure for stout handles."

  (name "Name-goes-here" 
    :short-doc "The title of the handle")

  (type 'stout-nil 
    :read-only 
    :short-doc 
    "The type of node the handle uses."
    :documentation
    "Node-types are symbols whose values have one of the types
stout-atom-node-type, stout-struct-node-type, stout-union-node-type,
or stout-array-node-type."))


;;; *** Node-type structures.

(defstruct stout-atom-node-type  
  "A pseudo-node-type with no internal structure.  
This node-type exists mostly to keep the code structure clean." )


(defstruct stout-struct-node-type
  "Node types that have members."

  (members nil
    :short-doc
    "The members of the node, which must be stout-handle-type.")
  
  (is-being-used nil

    :short-doc
    "Whether the type is in use, such as being printed."
    :documentation
    "This flag exists so that the printer routines et. al. don't get
into infinite loops if a type contains itself at any depth."

    ))



(defstruct stout-union-node-type
  "Node types that can be transformed into different types.

These are useful when some part of a structured document could take
different forms.  They allow a user to initially insert a less
specific node and later transform the node into some more specific
form."

  (transformations nil 
    :short-doc 
    "A list of types that nodes of this type can be transformed into."
    :documentation
    "All members of the list must be stout-handle-type"))



(defstruct stout-array-node-type
  "Array-like node types.

These are useful when some part of a structured document can be
repeated indefinitely many times."
  
  (element-type nil
    :short-doc
    "The element to be repeated, a stout-handle-type."))


;;; *** View information

;;This is unused for now.  When we have multiple views, it will be
;;important.

(defstruct stout-view
  "Information about a view"

  (root nil
    :short-doc
    "The object at the top of the view's hierarchy.")
  
  (name nil
    :short-doc
    "The name of the view."))


;;; *** Node information

(defstruct (stout-node-info (:type list))
  "Information about a single node"

  (type nil
    :short-doc
    "The type of the node."
    :documentation
"This must be of type stout-handle-type.")
  

  (ID-number nil
    :short-doc
    "The ID number for the node"
    :documentation
"This is usually unique within the document, but may not be if the
user has made changes.  Don't rely on this being stable.  It may
change when the view changes.")


  (participation nil
    :short-doc
    "A list indicating how the node participates in various views.")

  (visibility nil
    :short-doc
    "Whether the node is currently visible: t/nil.")
  
  (authorship nil
    :short-doc
    "Who created this node and when."
    :documentation
"If this is nil, the authorship is the same as that of the node above
it.  This feature is currently unused."
    )
  

  (application-info nil
    :short-doc
    "A slot set aside for specific applications."
    :documentation

    "Thru this slot, it is possible for outside elisp code to act on
data that's meaningfully related to stoutline structure."
    )

  (stored-text nil
    :short-doc
    "The body text of this node."
    :documentation

    "This is normally empty and ignored. It's used when storing a node
outside of a buffer."
    ))






;;; *** Node situation information

(defstruct stout-node-situ
  "Information about a single node's situation: Location, depth, etc."

  start-of-node
  end-of-header-string
  end-of-type-string 
  end-of-name-string
  
  (end-of-body nil
    :documentation
    "This is the same as the start of the next header.")

  (end-of-subtree nil
    :short-doc
    "Position at the beginning of the next node that's NOT deeper."
    :documentation
    "This slot is unused.")

  (depth nil
    :short-doc
    "How deep this node is in the outline hierarchy.")

  (object nil
    :short-doc
    "The object underlying this node."
    :documentation
    "This slot must be of type stout-node-info.")

  (at-buf-end nil
    :short-doc
    "Whether the node continues to the buffer's end."
    ))


;;; ** Functions:

;;; *** Helper functions

;;; **** To find positions

;;I didn't want to use outline.el's (outline-next-heading) because
;;that fiddles with outline-regexp, adding "\n" to the beginning.  IMO
;;that should be part of the regexp, not added by the code.

;;When p is exactly on the node-start, find the node that p is on, not
;;the next one.  re-search-forward has the correct behavior for this.

;;This essentially encapsulates re-search-forward so that we can feed
;;it a marker or position and have a marker returned, without changing
;;our current position.

(defun stout-start-of-next-node (p)
  "Return where the next node starts, or nil if there is no next node."

  (save-excursion
    (goto-char p)

    (let
      (	
	(result
	  (re-search-forward outline-regexp nil t)))

      ;;Important: Use match-beginning for the position, *not* what
      ;;re-search-forward returns, which is essentially (match-end 0).
      (if
	result
	(copy-marker (match-beginning 0))
	nil
	))))

  
;;; **** Type assertion
(defalias 'stout-appropriate-as-sym-p 'stringp)

;;; **** To classify nodes.

(defsubst stout-type-is-node-p (node-type)
  "Return t iff NODE-TYPE is any of the valid node types."

  (typecase node-type

    (stout-struct-node-type t ) 
    (stout-union-node-type  t )
    (stout-array-node-type  t )
    (stout-atom-node-type   t )
    (t nil)))


;;; **** To extract the type of a node.

(defsubst stout-get-node-type (node-type-name)
  "Get node-type from a name \(pseudo-symbol\) ."

  (cdr 
    (assoc 
      node-type-name
      stout-structuring-objects)))


(defun stout-object-get-node-type (node-data-object)
  "Get node-type from a lisp object.

NODE-DATA-OBJECT should be a lisp object that was read from a node
header."  

  (stout-get-node-type (symbol-name (car node-data-object))))





;;; **** To set text properties.

(defun stout-set-properties (str prop-list)
  "Set the string STR to have the properties in PROP-LIST.

This is useful for new text that's about to be inserted."
  
  (assert (stringp str))

  (let 
    ((len (length str))) 

    (set-text-properties 0 len prop-list str)))


;;; **** Default functions.

(defun stout-default-make-header-string (depth)
  (make-string depth ?* ))



;;; *** node-situ acquirement functions.

(defun stout-point-to-node-situ (p)

  "Return the situation of the node that marker P is in.

The return value is a stout-node-situ containing important buffer
positions, depth, and type.  

This is essentially a ctor for a stout-node-situ.  See the
documentation for stout-node-situ."

  ;;Assure that p is a marker.  If it was a position, change it into a
  ;;marker in the current buffer.
  (if (not ( markerp p))
    (progn
      (assert (integerp p))
      (setq p (copy-marker p))))
  
  ;;Don't change the user's position, if possible.
  ( save-excursion

    ;;Start at the given place.
    (set-buffer (marker-buffer p))
    (let

      ;;Declare the variables that we will use to fill the node we
      ;;will return.  It mite have made more sense to create a
      ;;stout-node-situ here and fill it as we go.

      ( start-of-node end-of-header-string end-of-type-string
	end-of-name-string end-of-body end-of-subtree object depth
	at-buf-end)


      ;;To avoid skipping our current position, we start one past our
      ;;current position unless that's impossible because we're at the
      ;;end of the buffer.  If we're at the end of the buffer, that
      ;;means we aren't overlooking a match that begins at p, because
      ;;nothing begins at p, so we're safe in not advancing p.
      (goto-char p)
      (if
	(not (eobp))
	(goto-char (1+ p)))
      
      
      ( let
	;;Make scratch variables
	( start-of-node-position found )

	
	(while 
	  (not found)
	  
	  ;;Search backwards. 
	  ( setq start-of-node-position
	    (re-search-backward outline-regexp nil t))
	  
	  ;;If we didn't even find a match, complain and give up.
	  (if
	    (null start-of-node-position)
	    (error "Position is before the first heading"))

	  ;;If the node-header we found is visible, we've found our
	  ;;node, otherwise let it repeat the search.  NB:
	  ;;re-search-backward has already moved the point to the new
	  ;;starting position, we need merely loop.
	  (setq found (outline-visible)))


	;;Get the longest possible match by searching-in-place with
	;;looking-at.  Otherwise we could get a match that didn't go
	;;past the starting position, IE, a match that was too short.
	(goto-char start-of-node-position)
	(looking-at outline-regexp)

	;;Store results into the appropriate objects.
	(setq start-of-node 
	  (copy-marker start-of-node-position))
	(setq end-of-header-string 
	  ( copy-marker (match-end 0))))
	 
	  

      ;;outline-level depends on the current position of point, which
      ;;should be at the start of a header line.  We reset point here
      ;;just in case.
      (goto-char start-of-node)
      (setq depth (funcall outline-level) )
      

      ;;Reading the node-type is a little involved because we use
      ;;"read" to do the work, which moves the marker itself.
      (let
	(
	  ;;Create a marker for the end of the type string, but place
	  ;;it at the beginning. "read" will move it to the end.  This
	  ;;must be a separate marker, or read will move the original
	  ;;marker as a side-effect.
	  (marker-in-type-string 
	    (copy-marker end-of-header-string t )))

	;;Read the node-type from marker-in-type-string, advancing
	;;that marker.
	(setq object (read marker-in-type-string ))
	
	;;We now have the proper position for the end of the type
	;;string.
	(setq end-of-type-string marker-in-type-string))
      

      ;;Move the point to where we will start searching for the rest
      ;;of the stuff.
      ( goto-char (marker-position end-of-type-string))

      
      ;;This throws its own errors, and those errors should be
      ;;impossible in a document using the default regexps.
      ( setq end-of-name-string 
	( copy-marker (re-search-forward stout-value-end-regxep)))


      (setq end-of-body (stout-start-of-next-node end-of-name-string))

      ;;If there was no next node, use the end of the buffer as the
      ;;end position. 
      (if
	(null end-of-body)

	(progn
	  (setq at-buf-end t)
	  (setq end-of-body (point-max-marker))))
      
    
      ;;We'd acquire end-of-subtree here if we ever want it.

    

      ;;Return a structure based on the acquired values.
      (make-stout-node-situ 
	:start-of-node        start-of-node 
	:end-of-header-string end-of-header-string 
	:end-of-type-string   end-of-type-string 
	:end-of-name-string   end-of-name-string
	:end-of-body          end-of-body
	:end-of-subtree       end-of-subtree
	:depth                depth
	:object	              object
	:at-buf-end           at-buf-end
	))))



;;; *** Insertion functions

(defun stout-insert-header ( handle depth &optional body-string )
  "Insert the header for a handle."

  ;;Assure that we have the rite type of arguments.
  (assert (stout-handle-type-p handle) )
  (assert (numberp depth) )

  (using-local-vars
     
    ;;Construct the appropriate strings: 

    ;;Outline-header string appropriate for the current depth.
    (local-var 
      header-string 
      (funcall stout-make-header-string depth ))
      
    ;;string representing the data object.  This initializes the
    ;;user-data to nil, but that can be set later. 
    ( local-var 
      handle-type-string 
      (format "%s"
	(make-stout-node-info 
	  :type (stout-handle-type-type handle)
	  :visibility t )))
      
    ;;Free-form name string.
    ( local-var handle-name-string (stout-handle-type-name handle) )
	    

    ;;Add properties to strings if desired
    (if stout-propertize-text 

      (progn

	(stout-set-properties header-string
	  stout-header-text-props )
 
	(stout-set-properties handle-type-string
	  stout-type-text-props)  
 
	(stout-set-properties handle-name-string
	  stout-name-text-props)
  
	(if body-string
	  (stout-set-properties body-string
	    stout-body-text-props))

	))


    ;;Map over the insertion format list...
    (mapcar
      (function 
	( lambda (ob)

	  ;;Insert each entry.
	  (let 
	    (
	      ( to-be-inserted
		( cond 
		  ((stringp ob) ob)
		  ((symbolp ob) (symbol-value ob))
		  (t "[Bad entry in stout-insert-format]" ))))

	    ;;If to-be-inserted is nil, ignore it.
	    (if 
	      to-be-inserted

	      ;;If the object is mistyped for insertion, complain and
	      ;;ignore it. 
	      (if
		(char-or-string-p to-be-inserted )
		(insert to-be-inserted)
		(message "Skipping an invalid format entry"))))))

      stout-insert-format)))
  


(defun stout-insert-struct-node-members ( node-type depth)
  "Insert the appropriate members for a struct node.

This function assumes we've already inserted the header and point is
positioned properly."

  ;;Check our arguments.
  (assert ( stout-struct-node-type-p node-type ))
  (assert ( numberp depth ))

  ;;We should not be in an infinite loop.  Any such loops should have
  ;;been cot elsewhere.
  (assert (not ( stout-struct-node-type-is-being-used node-type )))

  (unwind-protect

    (let
      ( (list-of-members ( stout-struct-node-type-members node-type )))

    
      ;;Set the recursion mark so we can detect whether this node-type
      ;;is being reused (IE, is part of a loop).
      (setf (stout-struct-node-type-is-being-used node-type) t)


      ;; Insert all members of the node
      (mapc
	( function
	  (lambda (handle )

	    ;;handle should be a stout-handle-type.  
	    (assert (stout-handle-type-p handle))
	    
	    ;;Insert that member (This may cause recursion, which is
	    ;;dealt with elsewhere).  
	    ( stout-insert-handle handle (1+ depth))))
	

	list-of-members))

    ;;Clear the recursion mark
    (setf (stout-struct-node-type-is-being-used node-type) nil)))



(defun stout-insert-handle ( handle depth &optional body-string)
  "Insert an entire handle."
  
  (assert (stout-handle-type-p handle) )
  (assert ( numberp depth ))

  ( using-local-vars

    ( local-var 
      node-type-name (stout-handle-type-type handle))

    ;;Test type for the new lookup scheme
    (assert (stout-appropriate-as-sym-p node-type-name))

    
    ( local-var 
      node-type (stout-get-node-type node-type-name) )
    
    (stout-insert-header handle depth body-string)
    
    (typecase node-type

      ;;If it is a struct type, insert its members.
      ( stout-struct-node-type 
	(stout-insert-struct-node-members node-type depth))

      ;;For other recognized types, do nothing
      ( stout-array-node-type ())
      ( stout-union-node-type ())
      ( stout-atom-node-type  ()) 

      ;;For unrecognized types, complain.
      (t (error "Unrecognized type"))

      )))



;;; *** Node-evolving functions

(defun stout-evolve-node ()
  "Evolve whichever node the point is currently in.

What it means to evolve a node depends on what type of node it is.

For union nodes, it means to transform it into one of the node-types
it can be transfomed into.

For array nodes, this means to insert one more element of whatever
type is appropriate.

For struct nodes, this has no meaning, since all their members should
have already been inserted.

For atom nodes, this has no meaning, since they have no structure."

  (interactive)

  ( using-local-vars 

    ( local-var 
      node-situ 
      (stout-point-to-node-situ (point-marker)))

    ( local-var 
      node-type-name 
      ( stout-node-situ-object node-situ ) )
 
    ( local-var 
      node-type 
      (stout-object-get-node-type node-type-name ) )

    (typecase node-type

      ( stout-array-node-type 
	( stout-array-add-element node-situ ))

      ( stout-union-node-type 
	(stout-union-transform node-situ))

      ( stout-struct-node-type ( error "Struct nodes can't evolve" ) )
      
      ( stout-atom-node-type ( error "Atom nodes can't evolve" ))

      ( t 
	( error 
	  "I don't recognize this node type, so I can't evolve it" ))
      
      )))


;;; **** Union-evolving functions


(defvar stout-union-transform-history nil 
  "The history list used by stout-get-union-transform.")


(defun stout-get-union-transform ( transformations )

  "Interactively select a type the union can transform into.

TRANSFORMATIONS is a list of all types the union can transform into."


  ( using-local-vars

    ;;This must be reconsidered after the change to alist
    (local-var my-alist
      (mapcar
	(function
	  (lambda (x)
	    ( cons (stout-handle-type-name x) x )))
    
	transformations))

    ( local-var my-type-name

      ( completing-read "Change into which type? " my-alist
	nil				;No predicate.
	t				;Require something that at
					;least expands to a match. 
	nil				;No known initial value.
	'stout-union-transform-history))
    
    ;;If we didn't abort, look up the type corresponding to the name.
    (if
      my-type-name
      (cdr (assoc my-type-name my-alist)))))



(defun stout-union-transform ( node-situ )
  "Transform a union type into another type.

The union node itself knows which types are possible transformations."


  ( using-local-vars

    ( local-var node-type-name (stout-node-situ-object    node-situ))
    ( local-var node-type    (stout-object-get-node-type node-type-name)) 

    
    ;;We must have a union type
    (assert (stout-union-node-type-p node-type))

    (local-var transformations 
      (stout-union-node-type-transformations node-type ) )

    ;;The union's transformations should be non-nil
    (assert transformations )

    ;;Pick a handle to transform into.
    (local-var chosen-handle (stout-get-union-transform transformations))
    
    ;;Proceed only if a handle was picked.
    (if
      chosen-handle
      ( using-local-vars
	

	;;Save the existing body.
	(local-var end-of-name-string
	  (stout-node-situ-end-of-name-string node-situ))
	
	(local-var start-of-node
	  (stout-node-situ-start-of-node node-situ))

	(local-var end-of-body
	  (stout-node-situ-end-of-body node-situ))

	(local-var body-string 
	  (buffer-substring end-of-name-string end-of-body ))

	( local-var depth 
	  (stout-node-situ-depth node-situ))

	;;Remove the existing node.
	(delete-region start-of-node end-of-body )

	;;Put point at the end of body, which is the only position
	;;remaining. 
	(goto-char (stout-node-situ-end-of-body node-situ))

	;;Insert the chosen handle, with the original body.
	(stout-insert-handle chosen-handle depth body-string )
    
	;;Put the mark at the beginning of the inserted stuff.
	(push-mark (stout-node-situ-end-of-body node-situ) )

	))))


;;; **** Array-evolving functions

(defun stout-array-add-element ( node-situ )
  "For an array node, add another element."

  ( using-local-vars

    ( local-var 
      node-type-name (stout-node-situ-object   node-situ))
    ( local-var 
      node-type    (stout-object-get-node-type   node-type-name)) 
    ( local-var 
      node-depth   (stout-node-situ-depth        node-situ))
    
    ;;We should have an array type
    (assert (stout-array-node-type-p node-type))

    ;;Get the handle to be inserted.
    (local-var 
      chosen-handle (stout-array-node-type-element-type node-type))

    
    ;;Put point at the end of body.  This looks funny if body is a lot
    ;;of blank lines, but it would be complex to fix it.
    (goto-char (stout-node-situ-end-of-body node-situ))

    ;;Insert the handle.
    (stout-insert-handle chosen-handle (1+ node-depth) nil )

    ;;Put the mark at the beginning of the inserted stuff.
    (push-mark (stout-node-situ-end-of-body node-situ) )

    ))




;; ** Sanity-check functions


(defun stout-sanity-check-node-type (node-type-name)
  "Return whether a node type has obvious problems.

For now, we only check struct nodes."

  (assert (stout-appropriate-as-sym-p node-type-name))

  (let
    ((node-type (stout-get-node-type node-type-name)))

    (typecase node-type

      ( stout-struct-node-type

	(if
	  (stout-struct-node-type-is-being-used node-type)

	  ;;If the member was already in use (IE, marked), we've hit a
	  ;;loop, so complain and don't recurse further.
	  (error "This type includes itself \(possibly indirectly\).")
		      
	  (stout-sanity-check-struct-node-type node-type)))

      ;;For other recognized types, do nothing (At least for now)
      ( stout-array-node-type ())
      ( stout-union-node-type ())
      ( stout-atom-node-type  ())

      ( t (error "Unrecognized type"))
      )))



;;If we want more speed, we mite make a mark in the type so that we
;;know which types have been already checked and can avoid rechecking
;;them.  This involves adding another handle to the node struct.  It
;;only makes sense to do this for struct nodes.

(defun stout-sanity-check-struct-node-type (node-type)
  "Return whether a struct node-type has obvious problems.

We check for self-inclusion at any depth, and we check that its'
lists, at any depth of recursion, all have type stout-handle-type.

For now, the error messages don't contain any dynamic information, but
that could change."

  ;;Check our arguments.
  (assert ( stout-struct-node-type-p node-type ))
 
  ;;We must not be in an infinite loop.  This should have been checked
  ;;elsewhere. 
  (assert (not ( stout-struct-node-type-is-being-used node-type )))

  (unwind-protect

    (let
      ( (list-of-members ( stout-struct-node-type-members node-type )))
	
      ;;Set the recursion mark so we can detect whether this node-type
      ;;is being reused (IE, is part of a loop).
      (setf (stout-struct-node-type-is-being-used node-type) t)


      ;;Map over all members of this node.
      (mapc
	( function
	  (lambda (handle )

	    (if 
	      (not (stout-handle-type-p handle))
	      (error 
		"All members of a struct node must be of type stout-handle-type")

	      ;;Recurse to check the member too.
	      (stout-sanity-check-node-type 
		(stout-handle-type-type handle)))))
      
	list-of-members))

    ;;Clear the recursion mark
    (setf (stout-struct-node-type-is-being-used node-type) nil)))


;;; *** To interactively insert nodes


(defvar stout-get-insertable-sym-history nil
  "History list for stout-get-insertable-sym.")

(defun stout-get-insertable-sym ()
  ""

  (using-local-vars

    ;;This throws an error if the user aborts.
    (local-var sym-name
      ( completing-read 
	"Type of node to insert? "  
	stout-structuring-objects
	nil
	t 
	nil
	'stout-get-insertable-sym-history))

    sym-name
    ))


(defun stout-insert-at-point (insert-type-sym)
  "Insert a stout object at point.

INSERT-TYPE-SYM must be a symbol for either a stout-handle-type object or
one of the node-types."

  (interactive (list (stout-get-insertable-sym)))

  (using-local-vars

    ;;Get the real object indicated by the name, if possible. 
    (local-var insert-type 
      (stout-get-node-type insert-type-sym))


    (cond

      ;;Interactively inserting a handle poses no special problems.
      ((stout-handle-type-p insert-type)
	(stout-insert-handle-at-point insert-type))
      
      ;;To insert a node instead of a handle, we interactively make an
      ;;anonymous handle that contains the node and insert that
      ;;instead.
      ((stout-type-is-node-p insert-type )
	(stout-insert-handle-at-point 
	  ( make-stout-handle-type :type insert-type-sym)))

      ;;If something else got thru, that's an error.
      (t
	(error "That object isn't a node or a handle.")))))



(defun stout-insert-handle-at-point (handle)
  "

HANDLE must be a stout-handle-type object."

  (assert (stout-handle-type-p handle))
  
  ( let
    ((depth 0))

    ;;Figure out depth at current point.  Anticipate errors.
    (condition-case err
      (using-local-vars

	(local-var node-situ 
	  (stout-point-to-node-situ (point-marker)))

	(setq depth 
	  ( stout-node-situ-depth node-situ ))
	
	;;Put point at a reasonable position for insertion, IE at the
	;;end of body.  This looks funny if body is a lot of blank
	;;lines, but it would be complex to fix it.
	(goto-char (stout-node-situ-end-of-body node-situ)))
      

      ;;On error, such as "before first heading", assume current depth
      ;;is 0.
      (error (setq depth 0)))
    
    (stout-insert-handle handle (1+ depth) )))


;;; *** To add properties to text when a file is loaded

(defun stout-propertize-buffer ()
  "Add the defined properties to stoutlined text in the current buffer."
  
  (interactive)

  ;;Paranoia.  The excursion is saved further in anyways.
  (save-excursion

    (let
      ( done 
	node-situ

	;;node-start-pos is initially before every node.
	(node-start-pos (point-min-marker)))
      
      ( while
	(not done)

	;;Advance NODE-START-POS to where the next node starts.  
	( setq node-start-pos (stout-start-of-next-node node-start-pos))

	(if
	  (null node-start-pos)

	  ;;If there was no "where the next node starts", we're done
	  ;;and we do nothing else during this iteration.
	  (setq done t)
	  
	  (progn

	    ;;Get the full position of the next node.  
	    (setq node-situ (stout-point-to-node-situ node-start-pos))

	    ;;Next iteration we'll be starting from where this node
	    ;;finishes, so remember where that is.  Store that
	    ;;position in NODE-START-POS.
	    (setq node-start-pos (stout-node-situ-end-of-body node-situ))

	    ;;Now we know the relevant positions/markers, so add the
	    ;;respective properties to the text between 'em.
	    (add-text-properties
	      (stout-node-situ-start-of-node         node-situ)
	      (stout-node-situ-end-of-header-string  node-situ)
	      stout-header-text-props)

	    (add-text-properties
	      (stout-node-situ-end-of-header-string  node-situ)
	      (stout-node-situ-end-of-type-string    node-situ)
	      stout-type-text-props)

	    (add-text-properties
	      (stout-node-situ-end-of-type-string    node-situ)
	      (stout-node-situ-end-of-name-string    node-situ)
	      stout-name-text-props)

	    (add-text-properties
	      (stout-node-situ-end-of-name-string    node-situ)
	      (stout-node-situ-end-of-body           node-situ)
	      stout-body-text-props)
	
	    ))))))


;;; *** To indent a subtree.

;;Future improvement: This or a variant of this could take "amount"
;;from the surrounding outline, such that the subtree becomes exactly
;;1 place deeper than its surroundings, rather than defaulting to 1.

(defun stout-indent-subtree (node-start-pos &optional amount)
  "Indent a subtree whose root NODE-START-POS is in, by AMOUNT.

When called interactively, \(point\) is used as NODE-START-POS.
AMOUNT defaults to 1."
  
  (interactive "d\np" )

  (if (null amount)
    (setq amount 1)
    (assert (numberp amount)))
  
  (let
    ( done node-situ target-depth)

    ( while
      (not done)

      ;;Get the full position of the node.  
      (setq node-situ (stout-point-to-node-situ node-start-pos))

      ;;Next iteration, we'll start searching from where this node
      ;;finishes, so remember where that is.  Store that position in
      ;;NODE-START-POS.
      (setq node-start-pos (stout-node-situ-end-of-body node-situ))


      (let
	((depth (stout-node-situ-depth node-situ)))

	(if

	  ;;If we don't know a target depth (IE, this is the first
	  ;;iteration of the loop) then this node is the root node, so
	  ;;use its depth as the target depth.
	  (null target-depth)
	  (progn
	    (setq target-depth depth)

	    ;;For the first iteration again: If we're indenting by a
	    ;;negative amount (unindenting) and the subtree just isn't
	    ;;as deep as that, stop before we ever modify anything.
	    (if
	      ( < (+ amount depth) 1)
	      (setq done t)))

	  ;;If the depth is less than the original node's depth, then
	  ;;we are out of the original subtree, so we stop and we
	  ;;don't alter this node.
	  (if
	    (<= depth target-depth)
	    (setq done t)))
	
	
      (if
	(not done)
	
	(using-local-vars

	  ;;Make a header of the new depth
	  (local-var header-string 
	    (funcall stout-make-header-string 
	      (+ depth amount)))
	  
	  ;;Set the new header's properties if appropriate.
	  (if stout-propertize-text 
	    (stout-set-properties header-string
	      stout-header-text-props ))

	  ;;Acquire the beginning and end of the existing header.
	  (local-var start-of-node
	    (stout-node-situ-start-of-node node-situ))

	  (local-var end-of-header-string
	    (stout-node-situ-end-of-header-string node-situ))

	  ;;Erase the old header
	  (delete-region start-of-node end-of-header-string )
	  
	  ;;Insert the new header in the old one's place.
	  (save-excursion
	    (goto-char start-of-node)
	    (insert header-string)))))
      

      ;;If this node is at the end of the buffer, then there aren't
      ;;any later nodes to indent so we're done.
      (if
	(stout-node-situ-at-buf-end node-situ)
	(setq done t)))))

;;; *** To handhold the user in defining a new structuring object.

;;Haven't done this yet, but it'd be convenient.

;;; *** To write out current structuring-objects as a file-local variable

;;Haven't done this yet, but it'd be convenient.


;;; ** User interface

;;; *** Keymap.  

;;This approach is a bit "rude", but I don't see how to add the
;;keystrokes elegantly, becase outline treats the keymaps differently
;;depending on whether it's a major or minor mode, and in any case
;;there are intervening prefix keys.
;;
;;So the easy/lazy thing to do is to alter outline mode's keymap,
;;which is what I do.

(define-key outline-mode-prefix-map "i" 'stout-insert-at-point)
(define-key outline-mode-prefix-map "\C-v" 'stout-evolve-node)
(define-key outline-mode-prefix-map "\C-z" 'stout-indent-subtree)


;;; *** Menubar (None for now)

;;We can set up a menubar thing easily enuff, probably borrowing
;;outline-mode-menu-bar-map, but I'll wait until I see an actual
;;desire for it.

;;; *** Mode miscellania

(defvar stout-mode-hook nil "Hooks for structured outline mode" )


;;; *** The mode proper

;;This code largely lets outline-mode do the work.
(defun stout-mode ()
  "Structured outline mode.

See stout.doc for details.  stout.doc should have been included with
the stout distribution.  

Important commands are:
\\<outline-mode-map>
\\[stout-insert-at-point] stout-insert-at-point Insert a node at point
\\[stout-evolve-node] stout-evolve-node     Evolve the node point is on
\\[stout-indent-subtree] stout-indent-subtree  Indent a subtree



More description should go here, but it hasn't."

  ;;We only handle outline major mode for now.
  (outline-mode)
  (setq mode-name "Stout")
  (setq major-mode 'stout-mode)
  
  (make-variable-buffer-local 'stout-structuring-objects)
  (make-variable-buffer-local 'stout-views)

  (let
    ((buffer-already-modified-p (buffer-modified-p)))
    
    (stout-propertize-buffer)

    ;;Adding properties doesn't change the buffer as far as we care, so
    ;;tell it it's unmodified unless there were other changes.
    (set-buffer-modified-p buffer-already-modified-p))

  (run-hooks 'stout-mode-hook))




;;; ** Epilog

(provide 'stout) 




;;Local Variables:
;;outline-regexp: "^;;; \\*+"
;;mode: outline-minor
;;change-log-default-name: "stout.changes"
;;End: 
;;
;;; stout.el ends here

